// free.c - free(1)
// procps utility to display free memory information
//
// All new, Robert Love <rml@tech9.net>             18 Nov 2002
// Original by Brian Edmonds and Rafal Maszkowski   14 Dec 1992
//
// This program is licensed under the GNU Library General Public License, v2
//
// Copyright 2003 Robert Love
// Copyright 2004 Albert Cahalan

#include "proc/sysinfo.h"
#include "proc/version.h"
//#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


const char help_message[] =
"usage: free [-b|-k|-m|-g|-h] [-l] [-o] [-t] [-s delay] [-c count] [-V]\n"
"  -b,-k,-m,-g show output in bytes, KB, MB, or GB\n"
"  -h human readable output (automatic unit scaling)\n"
"  -l show detailed low and high memory statistics\n"
"  -o use old format (no -/+buffers/cache line)\n"
"  -t display total for RAM + swap\n"
"  -s update every [delay] seconds\n"
"  -c update [count] times\n"
"  -a show available memory if exported by kernel (>80 characters per line)\n"
"  -V display version information and exit\n"
;

int shift = 10;
int human_readable = 0;


// Scale unit (-b, -k, -m, -g, -h)
char *S(unsigned long long X)
{
    const char *unit = "BKMGT";
    static char buf[41], buf2[11];
    int i;

    if (human_readable) {
	for (i = 0; unit[i] != '\0'; i++) {
	    switch (unit[i]) {
		case 'B':
		    if (4 >= snprintf(buf2, sizeof(buf2), "%llu%c", X << 10, unit[i])) {
			snprintf(buf, sizeof(buf), "%10s", buf2);
			return buf;
		    }
		    break;
		case 'K':
		    if (4 >= snprintf(buf2, sizeof(buf2), "%lld%c", X, unit[i])) {
			snprintf(buf, sizeof(buf), "%10s", buf2);
			return buf;
		    }
		    break;
		default:
		    if (4 >= snprintf(buf2, sizeof(buf2), "%.1f%c", (double)(X) / (1ULL << ((i-1)*10)), unit[i])) {
			snprintf(buf, sizeof(buf), "%10s", buf2);
			return buf;
		    }
		    if (4 >= snprintf(buf2, sizeof(buf2), "%lld%c", X >> ((i-1)*10), unit[i])) {
			snprintf(buf, sizeof(buf), "%10s", buf2);
			return buf;
		    }
	    }
        }
    } else {
	snprintf(buf, sizeof(buf), "%10Lu", (X << 10) >> shift);
    }

    return buf;
}


int main(int argc, char *argv[]){
    int i;
    int count = 0;
    int pause_length = 0;
    int show_high = 0;
    int show_total = 0;
    int show_available = 0;
    int old_fmt = 0;

    /* check startup flags */
    while( (i = getopt(argc, argv, "bkmghlotc:as:V") ) != -1 )
        switch (i) {
        case 'b': shift = 0;  break;
        case 'k': shift = 10; break;
        case 'm': shift = 20; break;
        case 'g': shift = 30; break;
        case 'h': human_readable = 1; break;
        case 'l': show_high = 1; break;
        case 'o': old_fmt = 1; break;
        case 't': show_total = 1; break;
        case 's': pause_length = 1000000 * atof(optarg); break;
        case 'c': count = strtoul(optarg, NULL, 10); break;
        case 'a': show_available = 1; break;
	case 'V': display_version(); exit(0);
        default:
            fwrite(help_message,1,strlen(help_message),stderr);
	    return 1;
    }

    do {
        meminfo();
        printf("             total       used       free     shared    buffers     cached");
        if (show_available) printf("  available");
        printf("\n");
        printf("Mem:    ");
        printf("%s ", S(kb_main_total));
        printf("%s ", S(kb_main_used));
        printf("%s ", S(kb_main_free));
        printf("%s ", S(kb_main_shared));
        printf("%s ", S(kb_main_buffers));
        printf("%s",  S(kb_main_cached));
        if (show_available) printf(" %s", S(kb_main_available));
        printf("\n");
        // Print low vs. high information, if the user requested it.
        // Note we check if low_total==0: if so, then this kernel does
        // not export the low and high stats.  Note we still want to
        // print the high info, even if it is zero.
        if (show_high) {
            printf("Low:    ");
            printf("%s ", S(kb_low_total));
            printf("%s ", S(kb_low_total - kb_low_free));
            printf("%s ", S(kb_low_free));
            printf("\n");
            printf("High:   ");
            printf("%s ", S(kb_high_total));
            printf("%s ", S(kb_high_total - kb_high_free));
            printf("%s ", S(kb_high_free));
            printf("\n");
        }
        if (!old_fmt) {
            unsigned KLONG buffers_plus_cached = kb_main_buffers + kb_main_cached;
            printf("-/+ buffers/cache: ");
            printf("%s ", S(kb_main_used - buffers_plus_cached));
            printf("%s ", S(kb_main_free + buffers_plus_cached));
            printf("\n");
        }
        printf("Swap:   ");
        printf("%s ", S(kb_swap_total));
        printf("%s ", S(kb_swap_used));
        printf("%s ", S(kb_swap_free));
        printf("\n");
        if (show_total) {
            printf("Total:  ");
            printf("%s ", S(kb_main_total + kb_swap_total));
            printf("%s ", S(kb_main_used  + kb_swap_used));
            printf("%s ", S(kb_main_free  + kb_swap_free));
            printf("\n");
        }
        if(pause_length){
	    fputc('\n', stdout);
	    fflush(stdout);
	    if (count != 1) usleep(pause_length);
	}
    } while(pause_length && --count);

    return 0;
}
